//HavokActor.cpp
#ifdef USE_HAVOKDEMO
#include <Demos/demos.h>
#endif
#include <Physics/Dynamics/World/hkpWorld.h>
#include <Physics/Dynamics/Phantom/hkpSimpleShapePhantom.h>
#include <Physics/Utilities/CharacterControl/CharacterProxy/hkpCharacterProxy.h>
#include <Physics/Dynamics/Entity/hkpRigidBody.h>
#include <Physics/Dynamics/Entity/hkpRigidBodyCinfo.h>
#include <Physics/Collide/Filter/Group/hkpGroupFilter.h>
#include <Physics/Collide/Shape/Convex/Capsule/hkpCapsuleShape.h>

#include <Physics/Utilities/CharacterControl/StateMachine/hkpCharacterContext.h>
#include <Physics/Utilities/CharacterControl/StateMachine/hkpCharacterState.h>
#include <Physics/Utilities/CharacterControl/StateMachine/hkpCharacterStateManager.h>
#include <Physics/Utilities/CharacterControl/StateMachine/OnGround/hkpCharacterStateOnGround.h>
#include <Physics/Utilities/CharacterControl/StateMachine/InAir/hkpCharacterStateInAir.h>
#include <Physics/Utilities/CharacterControl/StateMachine/Jumping/hkpCharacterStateJumping.h>
#include <Physics/Utilities/CharacterControl/StateMachine/Climbing/hkpCharacterStateClimbing.h>

#include "HavokActor.h"
#include "HavokManager.h"
#include <windows.h>

extern float Angle( float dx, float dy );

//---------------------------------------------------------------------------------
short GetKey( const short key )
{
	short retKey = 0;
	retKey = GetAsyncKeyState( key  ) & 0x8000;
	return retKey;
}
//---------------------------------------------------------------------------------
HavokActor::HavokActor()
{
	m_characterProxy = 0L;
#ifdef USE_ZUP
	m_up.set( 0.0f, 0.0f, 1.0f );;
	m_forward.set( 0.0f, -1.0f, 0.0f );
#else
	m_up.set( 0.0f, 1.0f, 0.0f );;
	m_forward.set( 0.0f, 0.0f, 1.0f );
#endif

	//
	// Create the Character state machine and context
	//
	{
		hkpCharacterState* state;
		hkpCharacterStateManager* manager = new hkpCharacterStateManager();

		state = new hkpCharacterStateOnGround();
		manager->registerState( state,	HK_CHARACTER_ON_GROUND);
		state->removeReference();

		state = new hkpCharacterStateInAir();
		manager->registerState( state,	HK_CHARACTER_IN_AIR);
		state->removeReference();

		state = new hkpCharacterStateJumping();
		manager->registerState( state,	HK_CHARACTER_JUMPING);
		state->removeReference();

		state = new hkpCharacterStateClimbing();
		manager->registerState( state,	HK_CHARACTER_CLIMBING);
		state->removeReference();

		m_characterContext = new hkpCharacterContext(manager, HK_CHARACTER_ON_GROUND);
		m_characterContext->setCharacterType( hkpCharacterContext::HK_CHARACTER_PROXY );
		m_characterContext->setFilterParameters(0.9f, 30, 400.0f);
		manager->removeReference();
	}


}
//---------------------------------------------------------------------------------
HavokActor::~HavokActor()
{
}
//---------------------------------------------------------------------------------
void HavokActor::CreateCapsuleProxy( const hkVector4& position, float strength, float mass )
{
	hkReal height = CHARACTER_SCALE_RATIO * 0.4f;
	hkReal radius = CHARACTER_SCALE_RATIO * 0.480f;

	hkpCapsuleShape* characterShape;
	const hkReal halfHeight = height*0.5f;

	hkVector4 vertexA(0, 0, halfHeight ); 
	hkVector4 vertexB(0, 0, -halfHeight ); 

	// Create a capsule to represent the character standing
	characterShape = new hkpCapsuleShape(vertexA, vertexB, radius);

	// Construct a Shape Phantom
	hkpSimpleShapePhantom* phantom = new hkpSimpleShapePhantom(
		characterShape, 
		hkTransform::getIdentity(), 
		hkpGroupFilter::calcFilterInfo(HavokManager::LAYER_ACTOR, 0 ));
	characterShape->removeReference();

	hkpWorld* pWorld = HavokManager::Instance()->GetWorld();
	pWorld->lock();

	// Add the phantom to the world
	pWorld->addPhantom(phantom);
	phantom->removeReference();

	// Construct a character proxy
	hkpCharacterProxyCinfo cpci;
	cpci.m_position.set(position(0), position(1), position(2));
	cpci.m_staticFriction = 0.0f;
	cpci.m_dynamicFriction = 1.0f;
	cpci.m_up.setNeg4( pWorld->getGravity() );
	cpci.m_up.normalize3();
	cpci.m_userPlanes = 4;
	cpci.m_maxCharacterSpeedForSolver = 1000.0f;
	cpci.m_maxSlope = 90.0f * HK_REAL_DEG_TO_RAD; //HK_REAL_PI / 3;

	cpci.m_shapePhantom = phantom;
	cpci.m_characterStrength = strength;
	cpci.m_characterMass = mass;
	this->m_characterProxy = new hkpCharacterProxy(cpci);
	

	//
	// Add in a custom listener
	//
	{
		//this->m_characterProxy->addCharacterProxyListener(new CharacterContactListener(this));

//		m_proxyListener = new GdcCharacterListener( UP, &m_options);
//		m_characterProxy->addCharacterProxyListener(m_proxyListener);
	}

	//this->m_currentTransform.setIdentity();
	//this->m_currentTransform.m_rotation.setAxisAngle( hkVector4(-1,0,0), HK_REAL_PI/2 );
	//this->m_currentTransform.m_translation = this->m_characterProxy->getPosition();

	pWorld->unlock();

	//this->m_horizontalDisplacement.setZero4();
	//this->m_verticalDisplacement = 0.0;

}
//---------------------------------------------------------------------------------
void HavokActor::DeleteProxy()
{
	hkpWorld* pWorld = HavokManager::Instance()->GetWorld();
	pWorld->lock();
	pWorld->removePhantom( this->m_characterProxy->getShapePhantom() );
	this->m_characterProxy->removeReference();
	pWorld->unlock();	

	delete m_characterContext;
	m_characterContext = HK_NULL;
}
//---------------------------------------------------------------------------------
void HavokActor::UpdateStep( float timestep ) 
{
#ifdef USE_HAVOKDIRECT_INPUT

    float fDX = 0.0f;
    float fDZ = 0.0f;
	float fStepWidth = CHARACTER_SCALE_RATIO * 0.5f;;

	bool jump = false;
	if( GetKey( VK_SPACE ) )
		jump = true;

	if( GetKey( 'P' ) )
	{
		//SetPosition( -25.229128f, -54.980015f, 15.400537f );
		SetPosition( -2.340109f, -156.680710f, 47.019413f ); 
	}

	hkpCharacterOutput output;
    if( GetKey( VK_UP ) )
    {
        fDZ += fStepWidth;
    }

	if( GetKey( VK_DOWN ) )
    {
        fDZ -= fStepWidth;
    }

    if( GetKey( VK_LEFT ) )
    {
        fDX -= fStepWidth;
    }

    if( GetKey( VK_RIGHT ) )
    {
        fDX += fStepWidth;
    }

	hkpWorld* pWorld = HavokManager::Instance()->GetWorld();
	if( pWorld == 0L )
		return;

	pWorld->lock();

	float down = 0;
	if( IsSupported() == false )
		down = -5;
	output.m_velocity = hkVector4( fDX, fDZ, down );

	//Apply the player character controller	
	{
		HK_TIMER_BEGIN( "simulate character", HK_NULL );

		// Feed output from state machine into character proxy
		m_characterProxy->setLinearVelocity( output.m_velocity );

		hkStepInfo si( hkTime(0.0f), hkTime(timestep) );
		m_characterProxy->integrate( si, pWorld->getGravity() );

		HK_TIMER_END();
	}

	pWorld->unlock();

#else
    float fDX = 0.0f;
    float fDZ = 0.0f;
	float fStepWidth = CHARACTER_SCALE_RATIO * 0.5f;;

	bool jump = false;
	if( GetKey( VK_SPACE ) )
		jump = true;

	if( GetKey( 'P' ) )
	{
		//SetPosition( -25.229128f, -54.980015f, 15.400537f );
		SetPosition( -2.340109f, -156.680710f, 47.019413f ); 
	}

    if( GetKey( VK_UP ) )
    {
        fDZ += fStepWidth;
    }

	if( GetKey( VK_DOWN ) )
    {
        fDZ -= fStepWidth;
    }

    if( GetKey( VK_LEFT ) )
    {
        fDX -= fStepWidth;
    }

    if( GetKey( VK_RIGHT ) )
    {
        fDX += fStepWidth;
    }

	hkpWorld* pWorld = HavokManager::Instance()->GetWorld();
	if( pWorld == 0L )
		return;

	pWorld->lock();

	hkpCharacterInput input;
	hkpCharacterOutput output;
	{
		input.m_inputLR = fDX;
		input.m_inputUD = fDZ;

		input.m_wantJump = jump;
		input.m_atLadder = false;

		input.m_up = m_up;
		input.m_forward = m_forward;

		hkQuaternion orient;
		orient.setAxisAngle( m_up, 0 );
		input.m_forward.setRotatedDir( orient, input.m_forward );

		hkVector4 characterGravity = m_up;
		float gravity = -16 * CHARACTER_SCALE_RATIO;
		characterGravity.mul4( hkVector4(  gravity, gravity, gravity ) );   
		input.m_characterGravity = characterGravity;

		input.m_stepInfo.m_deltaTime = timestep;
		input.m_stepInfo.m_invDeltaTime = 1.0f / timestep;
		input.m_velocity = m_characterProxy->getLinearVelocity();
		input.m_position = m_characterProxy->getPosition();

		hkVector4 down;	down.setNeg4( m_up );
		hkpSurfaceInfo ground;
		m_characterProxy->checkSupport(down, ground);

		input.m_surfaceInfo = ground;
		// Apply the character state machine
		{
			HK_TIMER_BEGIN( "update character state", HK_NULL );
			m_characterContext->update(input, output);
			HK_TIMER_END();
		}
	}


	//Apply the player character controller	
	{
		HK_TIMER_BEGIN( "simulate character", HK_NULL );

		// Feed output from state machine into character proxy
		m_characterProxy->setLinearVelocity(output.m_velocity);

		hkStepInfo si( hkTime(0.0f), hkTime(timestep) );
		m_characterProxy->integrate( si, pWorld->getGravity() );

		HK_TIMER_END();
	}


#if  0
	pWorld->lock();
#ifdef USE_ZUP
	hkVector4 desiredVelocity = hkVector4( fDX, fDZ, -10 );
#else
	hkVector4 desiredVelocity = hkVector4( fDX, -10, fDZ );
#endif
	this->m_characterProxy->setLinearVelocity( desiredVelocity );

	hkStepInfo si( pWorld->getCurrentTime(), pWorld->getCurrentTime() + timestep );
	m_characterProxy->integrate( si, pWorld->getGravity() );

#endif

	pWorld->unlock();

#endif
}
//---------------------------------------------------------------------------------
void HavokActor::SetPosition( float x, float y, float z )
{
	hkpWorld* pWorld = HavokManager::Instance()->GetWorld();
    pWorld->lock();
    pWorld->markForWrite();

	hkVector4 pos( x, y, z );
	m_characterProxy->setPosition( pos );

 	pWorld->unmarkForWrite();
    pWorld->unlock();
}
//--------------------------------------------------------------------------------
bool HavokActor::GetPosition( Vector3& pos )
{
	if( m_characterProxy == HK_NULL )
		return false;

	hkVector4 hkpos = m_characterProxy->getPosition();
	pos.x = hkpos(0);
	pos.y = hkpos(1);
	pos.z = hkpos(2);
	return true;
}
//---------------------------------------------------------------------------------
bool HavokActor::IsSupported()
{
	hkpSurfaceInfo ground;
	const hkVector4 down(0.0f, 0.0f, -1.0f);

	m_characterProxy->checkSupport( down, ground );
	bool bRes = ( ground.m_supportedState != hkpSurfaceInfo::UNSUPPORTED );
	return bRes;
}