//havok_config.cpp
#include "havok_config.h"

#include "framework/HavokContactListener.h"

//CheckContactPoint() 
#include <Physics/Collide/agent3/hkpagent3.h>
#include <Physics/Collide/agent3/machine/nn/hkpagentnntrack.h>
#include <Physics/dynamics/Collide/hkpsimpleconstraintcontactmgr.h>

#include "framework/HavokDefine.h"
#include "framework/HavokActor.h"
#include "framework/HavokCharacter.h"

hkpGroupFilter* g_groupFilter = NULL;

HavokActor* g_actor = NULL;
HavokCharacter* g_character = NULL;

void OutputDebugPrintf(char* fmt, ...) 
{
    char buffer[400];

    memset( buffer, 0, 400 );
    va_list argptr;
    int cnt;

    va_start(argptr, fmt);
    cnt = vsprintf_s(buffer, 400, fmt, argptr);
    va_end(argptr);

    OutputDebugString(buffer);
}

//HAvok  Ͼ  ȣѴ.

const float g_groundHalfHeight = 2.0f;
hkpRigidBody* g_groundBody = NULL;

//ڽ    
const float g_bx = 1.0f, g_by = 2.0f, g_bz = 1.0f;
const int g_nx = 1, g_ny = 1, g_nz = 1;
const int g_NUM_BODIES = g_nx * g_ny * g_nz;
hkpRigidBody *g_bodies[g_NUM_BODIES];
hkpRigidBody* g_sphereBody = NULL;
bool  g_bStopBall = false;

hkArray<hkContactPoint>		g_wallPoint;

void HavokErrorReportFunction(const char* s, void* errorReportObject) 
{
}


void SetEnableCollisionStaticBox( hkpWorld* pWorld, bool bFlag )
{
	// Replace filter
	pWorld->markForWrite();
	hkpGroupFilter* groupFilter = NULL;

	if( g_groupFilter == NULL )
	{
		g_groupFilter = new hkpGroupFilter();
		groupFilter = g_groupFilter;
		if( bFlag == true )
			groupFilter->enableCollisionsBetween( STATIC_LAYER, SPHERE_LAYER );
		else
			groupFilter->disableCollisionsBetween( STATIC_LAYER, SPHERE_LAYER );
		pWorld->setCollisionFilter( groupFilter, true);
		groupFilter->removeReference();
	}
	else
	{
		groupFilter = g_groupFilter;
		if( bFlag == true )
			groupFilter->enableCollisionsBetween( STATIC_LAYER, SPHERE_LAYER );
		else
			groupFilter->disableCollisionsBetween( STATIC_LAYER, SPHERE_LAYER );	
	}
	pWorld->unmarkForWrite();	
}

//position is center
hkpRigidBody* AddStaticBox( hkpWorld* pWorld , const hkVector4* position, const hkVector4*  width )
{
	const hkVector4& w = *width;
	const hkVector4 halfExtents( w(0)/2.0f, w(1)/2.0f, w(2)/2.0f );
	hkpShape* fixedSurfaceShape = new hkpBoxShape( halfExtents );

	//create rigid body information structure 
	hkpRigidBodyCinfo rigidBodyInfo;
	
	//MOTION_FIXED means static element in game scene
	rigidBodyInfo.m_mass = 0.0f;
	rigidBodyInfo.m_shape = fixedSurfaceShape;
	rigidBodyInfo.m_motionType = hkpMotion::MOTION_FIXED;
	rigidBodyInfo.m_position = *position;
	rigidBodyInfo.m_collisionFilterInfo = 
		hkpGroupFilter::calcFilterInfo(hkpGroupFilter::calcFilterInfo( STATIC_LAYER, 0));

	hkVector4 m_fAxis = hkVector4( 0, 0, 1 );
	//hkReal m_fAngle = -0.1f;
	//rigidBodyInfo.m_rotation = hkQuaternion( m_fAxis, m_fAngle);

	//create new rigid body with supplied info
	hkpRigidBody* pRigidBody = new hkpRigidBody( rigidBodyInfo );
	pRigidBody->addProperty( UID_WALL_PROPERTY, 7 );

	//add rigid body to physics world
	pWorld->lock();
	pWorld->addEntity( pRigidBody );

	//decerase reference counter for rigid body and shape
	pRigidBody->removeReference();
	fixedSurfaceShape->removeReference();

	pWorld->unlock();
	return pRigidBody;
}

hkpRigidBody* AddKeyFrameBox( hkpWorld* pWorld , const hkVector4* position, const hkVector4*  width )
{
	const hkVector4& w = *width;
	const hkVector4 halfExtents( w(0)/2.0f, w(1)/2.0f, w(2)/2.0f );
	hkpShape* fixedSurfaceShape = new hkpBoxShape( halfExtents );

	hkpRigidBodyCinfo rigidBodyInfo;
	
	rigidBodyInfo.m_mass = 0.0f;
	rigidBodyInfo.m_shape = fixedSurfaceShape;
	rigidBodyInfo.m_motionType = hkpMotion::MOTION_KEYFRAMED;
	rigidBodyInfo.m_position = *position;
	rigidBodyInfo.m_friction = 1.0f;
	rigidBodyInfo.m_collisionFilterInfo = 
		hkpGroupFilter::calcFilterInfo(hkpGroupFilter::calcFilterInfo( KEYFRAME_LAYER, 0));

	//create new rigid body with supplied info
	pWorld->lock();
	hkpRigidBody* pRigidBody = new hkpRigidBody( rigidBodyInfo );
	pRigidBody->addProperty( UID_WALL_PROPERTY, 8 );

	//add rigid body to physics world
	pWorld->addEntity( pRigidBody );

	//decerase reference counter for rigid body and shape
	pRigidBody->removeReference();
	fixedSurfaceShape->removeReference();

	pWorld->unlock();
	return pRigidBody;
}


bool MakeGround( hkpWorld* pWorld )
{
    const hkVector4 halfExtents(2000.0f, g_groundHalfHeight, 2000.0f);
    hkpBoxShape* groundShape = new hkpBoxShape(halfExtents);

    hkpRigidBodyCinfo bodyInfo;
    bodyInfo.m_mass = 0.0f; 
    bodyInfo.m_shape = groundShape;
    bodyInfo.m_motionType = hkpMotion::MOTION_FIXED;
	bodyInfo.m_collisionFilterInfo = 
		hkpGroupFilter::calcFilterInfo(hkpGroupFilter::calcFilterInfo( BACKGROUND_LAYER, 0));

    bodyInfo.m_position.set(0.0f, -g_groundHalfHeight, 0.0f);

    g_groundBody = new hkpRigidBody(bodyInfo);
	g_groundBody->addProperty( UID_WALL_PROPERTY, 3 );

    groundShape->removeReference();

	pWorld->markForWrite();
    pWorld->addEntity( g_groundBody );
	pWorld->unmarkForWrite();
    g_groundBody->removeReference(); 
	return true;
}

bool MakeBox( hkpWorld* pWorld )
{
    {
        const hkVector4 halfExtents(g_bx, g_by, g_bz);
        hkpShape* shape = new hkpBoxShape(halfExtents, 0.0f);

        hkpMassProperties massProperties;
        hkpInertiaTensorComputer::computeShapeVolumeMassProperties(shape, 5.0f, massProperties);

        hkpRigidBodyCinfo bodyInfo;
        bodyInfo.m_mass = massProperties.m_mass;
        bodyInfo.m_centerOfMass = massProperties.m_centerOfMass;
        bodyInfo.m_inertiaTensor = massProperties.m_inertiaTensor;
        bodyInfo.m_shape = shape;
        bodyInfo.m_motionType = hkpMotion::MOTION_BOX_INERTIA;

        int i = 0;
        for(int x = 0; x < g_nx; x++)
        for(int y = 0; y < g_ny; y++)
        for(int z = 0; z < g_nz; z++) {
            hkpRigidBody *body = new hkpRigidBody(bodyInfo);
            body->setPosition(hkVector4(x * g_bx * 2, 100.0f + y * g_by * 2, z * g_bz * 2));
            g_bodies[i] = body;
            pWorld->addEntity(body);

			HavokContactLister* listener = new HavokContactLister();
			body->addContactListener(listener);
			body->addEntityListener(listener);

            body->removeReference();
            i++;
        }
        shape->removeReference();
    }

	return true;
}

hkpRigidBody* AddDynamicShere( hkpWorld* pWorld , const hkVector4& position, float fRadius )
{
	const hkReal sphereMass = 30.0f;

	hkVector4 relPos( 0.0f, fRadius, 0.0f );

	hkpRigidBodyCinfo info;
	hkpMassProperties massProperties;
	hkpInertiaTensorComputer::computeSphereVolumeMassProperties( fRadius, sphereMass, massProperties );

	info.m_mass = massProperties.m_mass;
	info.m_centerOfMass  = massProperties.m_centerOfMass;
	info.m_inertiaTensor = massProperties.m_inertiaTensor;
	info.m_shape = new hkpSphereShape( fRadius );
	info.m_collisionFilterInfo = 
		hkpGroupFilter::calcFilterInfo(hkpGroupFilter::calcFilterInfo( SPHERE_LAYER, 0));
	info.m_position = position;
	//info.m_motionType  = hkpMotion::MOTION_BOX_INERTIA;
	info.m_motionType  = hkpMotion::MOTION_SPHERE_INERTIA;

	//info.m_qualityType = HK_COLLIDABLE_QUALITY_BULLET;

	hkpRigidBody* sphereRigidBody = new hkpRigidBody( info );

	pWorld->markForWrite();
	pWorld->addEntity( sphereRigidBody );
	
	HavokContactLister* listener = new HavokContactLister();
	sphereRigidBody->addContactListener(listener);
	
	sphereRigidBody->addEntityListener(listener);

	sphereRigidBody->removeReference();
	info.m_shape->removeReference();

	hkVector4 vel(  0.0f, 17.0f, 9.0f );
	sphereRigidBody->setLinearVelocity( vel );
	pWorld->unmarkForWrite();
	return sphereRigidBody;
}

void StopBall( hkpWorld* pWorld )
{
	pWorld->lock();
	//hkVector4 vel = g_sphereBody->getLinearVelocity();
	//g_sphereBody->setLinearVelocity( hkVector4( vel(0)/2.0f, vel(1)/2.0f, vel(2)/2.0f ) );

	g_sphereBody->setLinearVelocity( hkVector4( 0, 0 ,0 ) );
	g_sphereBody->setMotionType( hkpMotion::MOTION_FIXED );
	//߿ Mass  0 ϸ ״´.. ٸ ̻
	//g_sphereBody->setMass( 0.0f );

	pWorld->unlock();
}

void RestartBall( hkpWorld* pWorld, float x, float y, float z )
{
	pWorld->lock();
	g_sphereBody->setMotionType( hkpMotion::MOTION_SPHERE_INERTIA );
	g_sphereBody->setPosition( hkVector4( x, y, z ) );
	g_sphereBody->setLinearVelocity( hkVector4( 0.0f, 17.0f, 9.0f ) );
	pWorld->unlock();
}

float GetRadian( const D3DXVECTOR3& a, const D3DXVECTOR3& b )
{
    float cosAngle  = acosf( D3DXVec3Dot(&a, &b) / (D3DXVec3Length(&a) * D3DXVec3Length(&b)) );
    //cosAngle = D3DXToDegree( cosAngle );

    // z  ȴ.
    float angle =  (a.x * b.y - a.y * b.x > 0.0f) ? cosAngle : -cosAngle; 
    return angle;
}



bool IsCheckWall( const hkVector4& direction, const hkArray<hkContactPoint>& wallPoint )
{
	//浹 ġ ɹ 浹 ϴ ˻ϱ     Ѵ.
	hkVector4 vel = hkVector4( -direction(0), -direction(1), -direction(2) );
	vel.normalize3();

	hkReal desiredRad = HK_REAL_PI * 5.0f / 180.0f;

	int nSize = wallPoint.getSize();
	for( int n = 0; n < nSize; ++n )
	{
		hkVector4 nor = wallPoint[n].getNormal();
		hkVector4 wall = hkVector4( nor(0), nor(1), vel(2) );
		wall.normalize3();
		
		hkReal rad = GetRadian( D3DXVECTOR3( vel(0), vel(1), vel(2) ), D3DXVECTOR3( wall(0), wall(1), wall(2) ) );
		
		if( hkMath::fabs( rad ) < hkMath::fabs( desiredRad ) )
		{
			return true;
		}
	}
	return false;
}




#define FLOAT_EPSILON	1e-2f	
inline	float ABS(float a)		{ return (a < 0 ? -a : a); }

bool FloatEqual( float a, float b )
{
	float k = FLOAT_EPSILON;
	return fabs( a - b ) < FLOAT_EPSILON;
}


void CheckContactPoint( hkpWorld* pWorld, hkpRigidBody* pRigidBody, hkArray<hkContactPoint>& outPoint )
{
	outPoint.clear();

    hkStepInfo stepInfo;
	hkReal fTimestep = 1.0f / 60.0f;
    stepInfo.m_deltaTime = fTimestep;
    stepInfo.m_invDeltaTime = 1.0f / fTimestep;

	pWorld->markForWrite();
	hkpLinkedCollidable* coll = pRigidBody->getLinkedCollidable();
	hkArray<struct hkpLinkedCollidable::CollisionEntry> collisionEntriesTmp;
	coll->getCollisionEntriesSorted(collisionEntriesTmp);
	hkVector4 vel = pRigidBody->getLinearVelocity();
	hkVector4 angVel = pRigidBody->getAngularVelocity();
	const hkArray<struct hkpLinkedCollidable::CollisionEntry>& collisionEntries = collisionEntriesTmp;

	static int nLine = 0;
	int nUid = 0;
    for ( int i = 0; i < collisionEntries.getSize(); ++i )
    {
		hkpRigidBody* rb = hkpGetRigidBody( collisionEntries[i].m_partner );
		if( rb->hasProperty( UID_WALL_PROPERTY ) )
			nUid = rb->getProperty( UID_WALL_PROPERTY ).getInt();

        if ( rb != HK_NULL ) 
        {
			hkUint32 filter = rb->getCollisionFilterInfo();
			//if( filter > 0 )
			//	OutputDebugPrintf( "============׷ 浹 %d\n", filter );
            //if ( collisionEntries[i].m_agentEntry->m_contactMgr->m_type == hkpContactMgr::TYPE_SIMPLE_CONSTRAINT_CONTACT_MGR )
            {
                hkpSimpleConstraintContactMgr* mgr = (hkpSimpleConstraintContactMgr*)(collisionEntries[i].m_agentEntry->m_contactMgr);
                if( mgr->m_contactConstraintData.getNumContactPoints() > 0 )
                {
					hkContactPoint* contactPoints = mgr->m_contactConstraintData.m_atom->getContactPoints();
					
					for( int n = 0; n < mgr->m_contactConstraintData.getNumContactPoints(); ++n )
					{
						hkVector4 pos = contactPoints[n].getPosition();
						hkVector4 nor = contactPoints[n].getNormal();
						if( FloatEqual(vel(0), 0) && FloatEqual(vel(1), 0) && FloatEqual(vel(2), 0 )
							&& FloatEqual(angVel(0), 0) && FloatEqual(angVel(1), 0) && FloatEqual(angVel(2), 0) )
						{
							outPoint.pushBack( contactPoints[n] );
							//OutputDebugPrintf( "%d   %f, %f, %f \n", nUid, vel(0), vel(1), vel(2) );
						}
						else
						{
							outPoint.pushBack( contactPoints[n] );
							//OutputDebugPrintf( "%d : %d ӵ %f, %f, %f 浹  %.2f, %.2f, %.2f\n",
							//	nLine, nUid, vel(0), vel(1), vel(2),
							//	nor(0), nor(1), nor(2) );
						}
						++nLine;
					}
                }
            }
        }
    }

	pWorld->unmarkForWrite();
}

void CreateActor()
{
#ifdef USE_HAVOK_CHARACTER
	g_character = new HavokCharacter;
	g_character->Create( hkVector4( 0, 5, 0 ), 111 );
#else
	g_actor	= new HavokActor;
	g_actor->CreateCapsuleProxy( hkVector4( 0, 60, 0 ) );
#endif
}

void DeleteActor()
{
#ifdef USE_HAVOK_CHARACTER
	g_character->Destroy();
#else
	g_actor->DeleteProxy();
#endif
}

void UpdateActor()
{
#ifdef USE_HAVOK_CHARACTER
	g_character->Update( 1.f / 60.f );
#else
	g_actor->UpdateStep( 1.f / 60.f );
#endif
}

