﻿//main.cpp
#include <windows.h>
#include <tchar.h>
#include <d3d9.h>
#include <d3dx9.h>
#include <stdio.h>
#include <vector>

#include <Common/Base/hkBase.h>
#include <Physics/Dynamics/Entity/hkpRigidBody.h>
#include <Physics/Utilities/CharacterControl/CharacterProxy/hkpCharacterProxy.h>

#include "framework/HavokManager.h"
#include "framework/HavokLevel.h"
#include "framework/MakeRigidbody.h"
#include "obj_util/DxPolygon.h"
#include "util/HavokDebug.h"

#include "havok_config.h"
#include <Physics/Utilities/Constraint/Keyframe/hkpKeyFrameUtility.h>

#pragma comment(lib, "dxguid.lib")
#pragma comment(lib, "d3d9.lib")
#pragma comment(lib, "d3dx9.lib")

#pragma comment(lib, "winmm.lib")

TCHAR gName[100] = _T("Havok 시작");

hkpRigidBody* g_keyFrameRigidBox = NULL;

DxPolygon*    g_pObjModel = NULL;

struct DxRigidBody
{
	ID3DXMesh*		boxMesh;
	hkpRigidBody*	rigidBody;
};

std::vector< DxRigidBody > g_worldRigidBody;
LPDIRECT3D9 g_pD3D;
LPDIRECT3DDEVICE9 g_pD3DDev;


D3DMATERIAL9 g_staticBox_material = {
    {1.0f, 1.0f, 1.0f, 1.0f},
    {0.0f, 0.0f, 1.0f, 1.0f},
    {0.0f, 0.0f, 0.0f, 0.0f},
    {0.0f, 0.0f, 0.0f, 0.0f},
    10.0f
};

extern void OutputDebugPrintf(char* fmt, ...);


LRESULT CALLBACK WndProc(HWND hWnd, UINT mes, WPARAM wParam, LPARAM lParam){
    if(mes == WM_DESTROY) {PostQuitMessage(0); return 0;}
    return DefWindowProc(hWnd, mes, wParam, lParam);
}

void MakeStaticBox( D3DXVECTOR3 pos, D3DXVECTOR3 width )
{
	ID3DXMesh* box = NULL;
	D3DXCreateBox(g_pD3DDev, width.x, width.y, width.z, &box, 0);	
	
	hkpRigidBody* pRigidBody = 
		AddStaticBox( HavokManager::Instance()->GetWorld(), &hkVector4( pos.x, pos.y, pos.z ), &hkVector4( width.x, width.y, width.z ) );

	DxRigidBody dxRigidBody = { box, pRigidBody };
	g_worldRigidBody.push_back( dxRigidBody );
}

void MakeKeyFrameBox( D3DXVECTOR3 pos, D3DXVECTOR3 width )
{
	ID3DXMesh* box = NULL;
	D3DXCreateBox(g_pD3DDev, width.x, width.y, width.z, &box, 0);	
	
	hkpRigidBody* pRigidBody = 
		AddKeyFrameBox( HavokManager::Instance()->GetWorld(), &hkVector4( pos.x, pos.y, pos.z ), &hkVector4( width.x, width.y, width.z ) );

	DxRigidBody dxRigidBody = { box, pRigidBody };
	g_worldRigidBody.push_back( dxRigidBody );

	g_keyFrameRigidBox = pRigidBody;
}

void MakeDynamicSphere( D3DXVECTOR3 pos )
{
	ID3DXMesh* sphere = NULL;
	float radius = 1.5f;
	D3DXCreateSphere(g_pD3DDev, radius, 20, 20, &sphere, 0);


	hkpRigidBody* pRigidBody = 
		AddDynamicShere( HavokManager::Instance()->GetWorld(), hkVector4( pos.x, pos.y, pos.z ), radius );

	g_sphereBody = pRigidBody;

	DxRigidBody dxRigidBody = { sphere, pRigidBody };
	g_worldRigidBody.push_back( dxRigidBody );
}

void DrawStaticBox()
{
	//정적인 박스 그리기
	int nCount = (int)g_worldRigidBody.size();
	for( int n = 0; n < nCount; ++n )
	{
		const hkTransform &t = g_worldRigidBody[n].rigidBody->getTransform();
		D3DXMATRIX transMat(
			t(0,0), t(1,0), t(2,0), 0.0f,
			t(0,1), t(1,1), t(2,1), 0.0f,
			t(0,2), t(1,2), t(2,2), 0.0f,
			t(0,3), t(1,3), t(2,3), 1.0f);
		g_pD3DDev->SetTransform(D3DTS_WORLD, &transMat);
		g_pD3DDev->SetMaterial(&g_staticBox_material);

		g_worldRigidBody[n].boxMesh->DrawSubset(0);
	}
}

//#define FORCE_TRANSFORM   1

void OldMoveKeyFrameBox()
{
#ifdef   FORCE_TRANSFORM
	//최초 방법
	float time = (float)timeGetTime() / 1000.0f;
	const hkReal velAplitude = 10.0f; 
	const hkReal velOmega = 1.0f;

	hkVector4 newVelocity; 
	newVelocity = g_keyFrameRigidBox->getLinearVelocity();
	newVelocity(1) = velAplitude*cos(velOmega * time);
	
	g_keyFrameRigidBox->setLinearVelocity(newVelocity);
#endif

	static  float height = 0; 
	static bool	bPlus = true;
	float	step = 0.25f;	
	
	hkVector4 newPos;
	newPos = g_keyFrameRigidBox->getPosition();
	
	if( height >= 5 )
		bPlus = false;
	if( height <= -5 )
		bPlus = true;

	if( bPlus )
		height += step;
	else
		height -= step;

	newPos(1) = height;

#ifdef FORCE_TRANSFORM
	g_keyFrameRigidBox->setPosition( newPos );
#else
	hkVector4 pos = g_keyFrameRigidBox->getPosition();

	hkVector4 newVelocity = g_keyFrameRigidBox->getLinearVelocity();
	newVelocity(1) = ( newPos(1) - pos(1) ) * 60.0f;
	g_keyFrameRigidBox->setLinearVelocity( newVelocity );
#endif
}


void NewMoveKeyFrameBox()
{
	static  float height = 0; 
	static bool	bPlus = true;
	float	step = 0.25f;	
	
	hkpWorld* pWorld = HavokManager::Instance()->GetWorld();
	pWorld->lock();

	hkVector4 newPos;
	newPos = g_keyFrameRigidBox->getPosition();
	
	if( height >= 5 )
		bPlus = false;
	if( height <= -5 )
		bPlus = true;

	if( bPlus )
		height += step;
	else
		height -= step;

	newPos(1) = height;

	//hkpKeyFrameUtility::applyHardKeyFrame( )를 이용해 rigidBody 속도 설정
	hkQuaternion rotation;
	rotation.setIdentity();
	hkVector4 nextPos = newPos;
	hkpKeyFrameUtility::applyHardKeyFrame( nextPos, g_keyFrameRigidBox->getRotation(), 60.0f, g_keyFrameRigidBox );
	pWorld->unlock();
}


int APIENTRY _tWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow)
{
    //응용 프로그램 초기화
    MSG msg; HWND hWnd;
    WNDCLASSEX wcex ={sizeof(WNDCLASSEX), CS_HREDRAW | CS_VREDRAW, WndProc, 0, 0, hInstance, NULL, NULL, (HBRUSH)(COLOR_WINDOW+1), NULL, (TCHAR*)gName, NULL};
    if(!RegisterClassEx(&wcex))
        return 0;

    if(!(hWnd = CreateWindow(gName, gName, WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL)))
        return 0;

    // Direct3D 초기화
    if( !(g_pD3D = Direct3DCreate9( D3D_SDK_VERSION )) ) return 0;

    D3DPRESENT_PARAMETERS d3dpp = {0,0,D3DFMT_UNKNOWN,0,D3DMULTISAMPLE_NONE,0,D3DSWAPEFFECT_DISCARD,NULL,TRUE,0,D3DFMT_UNKNOWN,0,0}; 
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D24S8;
    d3dpp.MultiSampleType = D3DMULTISAMPLE_NONE;
    d3dpp.MultiSampleQuality = 0;
    d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_ONE;

    if( FAILED( g_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dpp, &g_pD3DDev ) ) )
    if( FAILED( g_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_SOFTWARE_VERTEXPROCESSING, &d3dpp, &g_pD3DDev ) ) ) {
        g_pD3D->Release();
        return 0;
    }

    ShowWindow(hWnd, nCmdShow);


    //////////////////////////////
    // Havok 초기화
	//InitHavok();
    //DDD
	HavokManager::CreateInstance( true );
	HavokLevel::CreateInstance();

	//비쥬얼 디버거 초기화
	HavokManager::Instance()->ActiveVisualDebugger();

	HavokDebug::CreateInstance();

    /////////////////////////////
    //바닥을 만든다.
	MakeGround( HavokManager::Instance()->GetWorld() );
	MakeStaticBox( D3DXVECTOR3( 0, 5, 15 ), D3DXVECTOR3( 30, 10, 2 ) ); 
	MakeStaticBox( D3DXVECTOR3( -20, 0, 0 ), D3DXVECTOR3( 2, 10, 30 ) ); 
	MakeKeyFrameBox( D3DXVECTOR3( 0, 12, 0 ), D3DXVECTOR3( 10, 5, 10 ) ); 

	//OBJ 파일 열기
	//g_pObjModel = new DxPolygon;
	if( g_pObjModel )
		g_pObjModel->LoadObj( "collision.obj" );
	//DDD
	//HavokLevel::Instance()->AddWorld( "collision.hkx" );
	HavokLevel::Instance()->AddPhysicsSystem( "move_collision.hkx" );

	//동적인 구 만들기
	MakeDynamicSphere( D3DXVECTOR3( 0, 7, -20 ) );

	CreateActor();	

    //Box를 만들어 보자
	//MakeBox( HavokManager::Instance()->GetWorld() );

    //바닥 DX 메쉬 만들기
    ID3DXMesh *floor, *box;
    D3DXCreateBox(g_pD3DDev, g_bx * 2, g_by * 2,g_bz * 2, &box, 0);

	//동적인 DX 메쉬 만들기
    D3DXCreateBox(g_pD3DDev, 4000.0f, g_groundHalfHeight * 2.0f, 4000.0f, &floor, 0);

    //ID3DXFont 구성 요소 생성
    int fontsize = 24;
    D3DXFONT_DESC lf = {fontsize, 0, 0, 1, 0, HANGEUL_CHARSET, OUT_TT_ONLY_PRECIS, PROOF_QUALITY, FIXED_PITCH | FF_MODERN, _T("바탕")};
    LPD3DXFONT pD3DFont;
    if(FAILED(D3DXCreateFontIndirect(g_pD3DDev, &lf, &pD3DFont))) {
        g_pD3DDev->Release(); g_pD3D->Release();
        return 0;
    }

    //변환 행렬
    D3DXMATRIX view, proj;
    D3DXMatrixPerspectiveFovLH( &proj, D3DXToRadian(45), 640.0f/480.0f, 1.0f, 10000.0f);
    g_pD3DDev->SetTransform(D3DTS_PROJECTION, &proj);

    //라이트
    D3DLIGHT9 light;
    ZeroMemory(&light, sizeof(D3DLIGHT9));
    light.Direction = D3DXVECTOR3(-1, -2, 0.5);
    light.Type = D3DLIGHT_DIRECTIONAL; 
    light.Diffuse.r = 0.7f;
    light.Diffuse.g = 0.7f;
    light.Diffuse.b = 0.7f;
    light.Diffuse.a = 1.0f;
    g_pD3DDev->SetLight( 0, &light );
    g_pD3DDev->LightEnable( 0, true );
    g_pD3DDev->SetRenderState( D3DRS_LIGHTING, TRUE );
    g_pD3DDev->SetRenderState( D3DRS_AMBIENT, 0x00808080 );
	g_pD3DDev->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE );

    //재질
    D3DMATERIAL9 ground_material = {
        {1.0f, 1.0f, 1.0f, 1.0f},
        {0.2f, 0.2f, 0.2f, 1.0f},
        {0.0f, 0.0f, 0.0f, 0.0f},
        {0.0f, 0.0f, 0.0f, 0.0f},
        10.0f
    };
    D3DMATERIAL9 cube_material = {
        {0.7f, 1.0f, 0.7f, 1.0f},
        {0.2f, 0.2f, 0.2f, 1.0f},
        {0.0f, 0.0f, 0.0f, 0.0f},
        {0.0f, 0.0f, 0.0f, 0.0f},
        10.0f
    };


    //메시지 루프
    float tm = 0.0f;
    do
	{
        if( PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) )
		{
            DispatchMessage(&msg);
		}
		else 
		{
			hkpWorld* pWorld = HavokManager::Instance()->GetWorld();
			if( ::GetAsyncKeyState(VK_RETURN) & 0x8000f )
			{
				RestartBall(pWorld, 0, 7, -20 );
			}
			if( ::GetAsyncKeyState('1') & 0x8000f )
			{
				g_bStopBall = !g_bStopBall;
			}
			if( ::GetAsyncKeyState('2') & 0x8000f )
			{
				SetEnableCollisionStaticBox( pWorld, false );
			}
			if( ::GetAsyncKeyState('3') & 0x8000f )
			{
				SetEnableCollisionStaticBox( pWorld, true );
			}
			if( ::GetAsyncKeyState('A') & 0x8000f )
			{
				RestartBall( pWorld, -10, 10, -20 );
			}
			if( ::GetAsyncKeyState('S') & 0x8000f )
			{
				RestartBall( pWorld, 5, 6, -20 );
			}


			tm += 1.0f/60.0f;
			hkVector4  pos, t; 

			//OldMoveKeyFrameBox();
			//OldMoveKeyFrameBox( ), NewMoveKeyFrameBox( ) 함수의 결과는 같다.
			NewMoveKeyFrameBox();
			UpdateActor();

			HavokManager::Instance()->Update();
			CheckContactPoint( pWorld, g_sphereBody, g_wallPoint );
			
			static int nLine = 0;
			pos = g_keyFrameRigidBox->getPosition();
			hkVector4  vel = g_keyFrameRigidBox->getLinearVelocity();
			//OutputDebugPrintf( "%d : 위치 %.2f, %.2f, %.2f, 속도 : %.2f, %.2f, %.2f\n", nLine, pos(0), pos(1), pos(2), vel(0), vel(1), vel(2) );
			nLine++;
			{	
				D3DXMatrixLookAtLH(&view, &D3DXVECTOR3(0.0f, 50.0f, -25.0f), &D3DXVECTOR3( 0, 0, 0), &D3DXVECTOR3(0.0f, 1.0f, 0.0f));
				g_pD3DDev->SetTransform(D3DTS_VIEW, &view);
			}

			char strDisplay[128];
			sprintf_s( strDisplay, 128, "'2'키 : 스태틱 박스 공 통과, '3'키: 스택틱 박스 충돌 %5.1f sec.\n", tm );

			//Direct3D 처리
			g_pD3DDev->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER | D3DCLEAR_STENCIL, D3DCOLOR_XRGB(0,0,255), 1.0f, 0 );
			g_pD3DDev->BeginScene();

			//바닥 그리기
			{
				const hkTransform &t = g_groundBody->getTransform();
				D3DXMATRIX transMat(
					t(0,0), t(1,0), t(2,0), 0.0f,
					t(0,1), t(1,1), t(2,1), 0.0f,
					t(0,2), t(1,2), t(2,2), 0.0f,
					t(0,3), t(1,3), t(2,3), 1.0f);
				g_pD3DDev->SetTransform(D3DTS_WORLD, &transMat);
				g_pD3DDev->SetMaterial(&ground_material);

				floor->DrawSubset(0);
			}

			//큐브 그리기
			{
				for (int i = 0; i < g_NUM_BODIES; i++) 
				{
					if( g_bodies[i] == NULL )
						break;
					const hkTransform &t = g_bodies[i]->getTransform();
					D3DXMATRIX transMat(
						t(0,0), t(1,0), t(2,0), 0.0f,
						t(0,1), t(1,1), t(2,1), 0.0f,
						t(0,2), t(1,2), t(2,2), 0.0f,
						t(0,3), t(1,3), t(2,3), 1.0f);
					g_pD3DDev->SetTransform(D3DTS_WORLD, &transMat);
					g_pD3DDev->SetMaterial(&cube_material);
					box->DrawSubset(0);
				}
			}

			DrawStaticBox();

			if( g_pObjModel )
				g_pObjModel->Render( g_pD3DDev );
			if( HavokDebug::Instance() )
				HavokDebug::Instance()->Render();

			//위치 그리기
			RECT rc = {0, 0, 0, 0};
			pD3DFont->DrawTextA(NULL, strDisplay, -1, &rc, DT_LEFT | DT_SINGLELINE | DT_NOCLIP, 0xffffffff);

			g_pD3DDev->EndScene();
			g_pD3DDev->Present( NULL, NULL, NULL, NULL );
		}
    } while(msg.message != WM_QUIT);

	if( g_pObjModel )
		delete g_pObjModel;

	DeleteActor();
	g_wallPoint.clearAndDeallocate();

    floor->Release();
    box->Release();

	int nCount = (int)g_worldRigidBody.size();
	for( int n = 0; n < nCount; ++n )
	{
		g_worldRigidBody[n].boxMesh->Release();
	}

	HavokDebug::ReleaseInstance();
	HavokLevel::Instance()->ReleaseInstance();
	HavokManager::Instance()->ReleaseInstance();

    pD3DFont->Release();
    g_pD3DDev->Release();
    g_pD3D->Release();

    return 0;
}

