//DxVertex.cpp
#include <windows.h>
#include <tchar.h>
#include <d3d9.h>
#include <d3dx9.h>
#include <stdio.h>
#include <vector>

#include "DxMesh.h"

const DWORD DxVertex::FVF = D3DFVF_XYZ | D3DFVF_NORMAL;

extern LPDIRECT3DDEVICE9 g_pD3DDev;

//----------------------------------------------------------------------
DxMesh::DxMesh()
{
	m_bWireFrame = false;
	m_indexNum = m_vertexNum = 0;
	m_vb = NULL;
	m_ib = NULL;
	m_pMesh = NULL;

	D3DXMatrixIdentity( &m_trans );
}
//----------------------------------------------------------------------
DxMesh::~DxMesh()
{
	Release();
}
//----------------------------------------------------------------------
void DxMesh::Release()
{
	if( m_pMesh )
	{
		m_pMesh->Release();
		m_pMesh = NULL;
	}
}
//----------------------------------------------------------------------
bool DxMesh::CreateVertexBuffer()
{
	HRESULT hr = 0;
	hr = D3DXCreateMeshFVF(
		m_indexNum / 3,
		m_vertexNum,
		D3DXMESH_MANAGED,
		DxVertex::FVF,
		g_pD3DDev,
		&m_pMesh);

	if(FAILED(hr))
		return false;

	DWORD dbgNumVertex = m_pMesh->GetNumVertices();
	DWORD dbhNumFace = m_pMesh->GetNumFaces();
	DxVertex* v = 0;
	m_pMesh->LockVertexBuffer(0, (void**)&v);
	for( DWORD n = 0; n < m_vertexNum; ++n )
		v[n] = m_vb[n];
	m_pMesh->UnlockVertexBuffer();

	WORD* pIndex = 0;
	m_pMesh->LockIndexBuffer(0, (void**)&pIndex );
	for( DWORD n = 0; n < m_indexNum; n += 3 )
	{
		//DDD
#if   1
		pIndex[n + 0] = m_ib[n + 0].index;
		pIndex[n + 1] = m_ib[n + 1].index;
		pIndex[n + 2] = m_ib[n + 2].index;
#else
		pIndex[n + 2] = m_ib[n + 0];
		pIndex[n + 1] = m_ib[n + 1];
		pIndex[n + 0] = m_ib[n + 2];
#endif
	}
	m_pMesh->UnlockIndexBuffer();
	return true;
}
//----------------------------------------------------------------------
void DxMesh::Render()
{
	if( m_bWireFrame )
	{
		g_pD3DDev->SetRenderState( D3DRS_LIGHTING, FALSE );
		g_pD3DDev->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );
	}

	static D3DMATERIAL9 material = {
			{1.0f, 1.0f, 1.0f, 1.0f},
			{0.0f, 0.0f, 1.0f, 1.0f},
			{0.0f, 0.0f, 0.0f, 0.0f},
			{0.0f, 0.0f, 0.0f, 0.0f},
			10.0f  };

	if( m_pMesh == NULL )
		CreateVertexBuffer();

	DWORD numSubsets = 0;
	m_pMesh->GetAttributeTable(0,  &numSubsets);

	g_pD3DDev->SetTransform(D3DTS_WORLD, &m_trans);	

	g_pD3DDev->SetMaterial(&material);
	if( numSubsets == 0 )
	{
		m_pMesh->DrawSubset( 0 );
	}
	else
	{
		for(DWORD n = 0; n < numSubsets; ++n)
		{
			m_pMesh->DrawSubset( n );
		}
	}

	if( m_bWireFrame )
	{
		g_pD3DDev->SetRenderState( D3DRS_LIGHTING, TRUE );
		g_pD3DDev->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
	}
}
//----------------------------------------------------------------------