//HavokDebug
#include "HavokDebug.h"

#include <Physics/Dynamics/Entity/hkpRigidBody.h>
#include <Common/SceneData/Scene/hkxScene.h>
#include <Common/SceneData/Scene/hkxSceneUtils.h>
#include <Common/SceneData/Mesh/hkxMeshSection.h>
#include <Common/SceneData/Mesh/hkxVertexDescription.h>
#include <Common/SceneData/Mesh/hkxVertexBuffer.h>
#include <Common/SceneData/Mesh/hkxIndexBuffer.h>
#include <Physics/Utilities/Serialize/hkpPhysicsData.h>

//׿ ޽ ؼ
#include <Common/Visualize/Shape/hkDisplayGeometry.h>
#include <Physics/Utilities/VisualDebugger/Viewer/hkpShapeDisplayBuilder.h>
#include <Common/Base/Types/Geometry/hkGeometry.h>

#include <windows.h>
#include <tchar.h>
#include <d3d9.h>
#include <d3dx9.h>
#include <stdio.h>
#include <vector>
#include "../framework/PhysicsDefine.h"
#include "DxMesh.h"

extern void OutputDebugPrintf(char* fmt, ...); 

HavokDebug* HavokDebug::m_pInstance = NULL;

static void ComputeNormal( D3DXVECTOR3* out, D3DXVECTOR3* p0, D3DXVECTOR3* p1, D3DXVECTOR3* p2 )
{
    D3DXVECTOR3 u = *p1 - *p0; 
    D3DXVECTOR3 v = *p2 - *p0;
 
      D3DXVec3Cross(out, &u, &v);
      D3DXVec3Normalize(out, out);
}
//-----------------------------------------------------------------------------
bool HavokDebug::CreateInstance()
{
	if( m_pInstance )
		return true;
	m_pInstance = new HavokDebug();
	m_pInstance->Init();
	return true;
}
//-----------------------------------------------------------------------------
void HavokDebug::ReleaseInstance()
{
	if( m_pInstance )
	{
		m_pInstance->UnInit();
		delete m_pInstance;
		m_pInstance = NULL;
	}
}
//-----------------------------------------------------------------------------
HavokDebug* HavokDebug::Instance()
{
	return m_pInstance;
}
//-----------------------------------------------------------------------------
HavokDebug::HavokDebug()
{

}
//-----------------------------------------------------------------------------
HavokDebug::~HavokDebug()
{

}
//-----------------------------------------------------------------------------
bool HavokDebug::Init()
{
	return true;
}
//-----------------------------------------------------------------------------
void HavokDebug::UnInit()
{
	int nSize = (int)m_meshArr.size();
	for( int n = 0; n < nSize; ++n )
	{
		MeshInfo& info = m_meshArr[n];
		DxMesh* pMesh = info.dxMesh;
		delete pMesh;
	}
	m_meshArr.clear();
}
//-----------------------------------------------------------------------------
void HavokDebug::MakeMesh( hkpShape* pShape, hkpRigidBody* pRigidBody )
{
    hkArray<hkDisplayGeometry*> displayGeometry;
    hkpShapeDisplayBuilder::hkpShapeDisplayBuilderEnvironment env;
    hkpShapeDisplayBuilder builder(env);

    builder.buildDisplayGeometries( pShape, displayGeometry );

	int nSize = displayGeometry.getSize();
	for( int n = 0; n < nSize; ++n )
	{
		displayGeometry[n]->buildGeometry();

		const hkGeometry* geom = displayGeometry[n]->getGeometry();

		int numTriangles = geom->m_triangles.getSize();
		DWORD numIndices = numTriangles * 3;

		DxIndex* indexPtr = new DxIndex[numIndices];
		DxIndex* curIndex = indexPtr;

		int numVerts = geom->m_vertices.getSize();
		DxVertex* vertexData = new DxVertex[numVerts];

		for(int currentVertex = 0; currentVertex < numVerts; ++currentVertex)
		{
			// Get the vertex data for each point
			const hkVector4& pos = geom->m_vertices[ currentVertex ];
			vertexData[currentVertex].position = D3DXVECTOR3(pos(0), pos(1), pos(2));
			vertexData[currentVertex].normal = D3DXVECTOR3(0, 1, 2 );
		}

		for( int currentTriangle = 0; currentTriangle < numTriangles; ++currentTriangle)
		{
			const hkGeometry::Triangle& triIndices = geom->m_triangles[ currentTriangle ];

			curIndex[0].index = triIndices.m_a;
			curIndex[1].index = triIndices.m_b;
			curIndex[2].index = triIndices.m_c;
			curIndex += 3;

			//ְ Ȯ   Ѵ.
			D3DXVECTOR3 out;
			ComputeNormal( &out, 
				&(vertexData[ triIndices.m_a ].position),
				&(vertexData[ triIndices.m_b ].position),	
				&(vertexData[ triIndices.m_c ].position ) );

			vertexData[triIndices.m_a ].normal = out;
			vertexData[triIndices.m_b ].normal = out;
			vertexData[triIndices.m_c ].normal = out;

			displayGeometry[n]->releaseGeometry();
		}

		//Create DxMesh
		DxMesh* dxMesh = new DxMesh;
		dxMesh->SetIndexBuffer( numIndices, indexPtr );
		dxMesh->SetVertexBuffer( numVerts, vertexData );
		//dxMesh->SetWireFrame( true );

		MeshInfo info;
		info.dxMesh = dxMesh;
		info.refRigidBody = pRigidBody;
		m_meshArr.push_back( info );
	}
}
//-----------------------------------------------------------------------------
void HavokDebug::MakeMesh( hkxScene* scenePtr, hkpPhysicsData* physics )
{
	if ( scenePtr == HK_NULL || (scenePtr->m_rootNode == HK_NULL) )
	{
		HK_ASSERT(0xa54323c0, "No scene data (or scene data root hkxNode) found. Can't continue.");
		return;
	}

#if   1
	//庰 ̵, ȸ ִ° 
	hkxNode* rootNode = scenePtr->m_rootNode;
	MakeStaticMesh( scenePtr, physics, rootNode );
#else
	//庰 ̵, ȸ ° 
	for( int nMesh = 0; nMesh < scenePtr->m_meshes.getSize(); nMesh++ )
	{
		hkxMesh* originalMesh = scenePtr->m_meshes[nMesh];

		for( int nSection = 0; nSection < originalMesh->m_sections.getSize(); ++nSection )
		{
			hkxMeshSection* section = originalMesh->m_sections[nSection];
			MakeVertex( section, worldTransform );
		}
	}
#endif
}
//-----------------------------------------------------------------------------
void HavokDebug::MakeStaticMesh( hkxScene* scenePtr, hkpPhysicsData* physics, hkxNode* pNode )
{
	BuildMesh( scenePtr, physics, pNode, reinterpret_cast<hkxMesh*>(pNode->m_object.val()));

	// Go through each of our children now
	for(int i = 0; i < pNode->m_children.getSize(); ++i)
	{
		 MakeStaticMesh( scenePtr, physics, pNode->m_children[i] );
	}

}
//-----------------------------------------------------------------------------
void HavokDebug::BuildMesh( hkxScene* scenePtr, hkpPhysicsData* physics, hkxNode* pNode, hkxMesh* pMesh )
{
	if( pMesh == HK_NULL )
		return;

	// Get the orientation of the object
	float transformData[16];
	hkMatrix4 mat;
	hkQsTransform transform;
	scenePtr->getWorldFromNodeTransform( pNode, mat);
	mat.get4x4ColumnMajor( transformData );
	transform.set4x4ColumnMajor( transformData );

	// ̵
	const char* nodeName = pNode->m_name;
	hkpRigidBody* body =  physics->findRigidBodyByName( nodeName );
	if( body )
	{
		//
		const hkTransform& orgTrans = body->getTransform();
		//3dsMax ǥ ýۿ  ̵, ȸ Űܵ  
		//Pivot ̵, ȸ ϰ Ǹ VisualDebugger ̻ϰ 
	}

	for( int nSection = 0; nSection < pMesh->m_sections.getSize(); ++nSection )
	{
		hkxMeshSection* section = pMesh->m_sections[nSection];
		MakeVertex( section, mat );
	}
}
//-----------------------------------------------------------------------------
void HavokDebug::MakeVertex( hkxMeshSection* section, const hkMatrix4& worldTransform )
{
	float minX = 1000000.0f, maxX = -1000000.0f,
			minY = 1000000.0f, maxY = -1000000.0f,
			minZ = 1000000.0f, maxZ = -1000000.0f;

	hkxVertexBuffer* vb = section->m_vertexBuffer;
	hkxIndexBuffer* ib = section->m_indexBuffers[0];
	hkxMaterial* material = section->m_material;

	// Get the vertex description
	const hkxVertexDescription& vertexDesc = vb->getVertexDesc();

	// Get the declarations for each of the pieces of data we want
	const hkxVertexDescription::ElementDecl* posDecl =
		vertexDesc.getElementDecl(hkxVertexDescription::HKX_DU_POSITION, 0);
	const hkxVertexDescription::ElementDecl* normDecl =
		vertexDesc.getElementDecl(hkxVertexDescription::HKX_DU_NORMAL, 0);
	const hkxVertexDescription::ElementDecl* tanDecl =
		vertexDesc.getElementDecl(hkxVertexDescription::HKX_DU_TANGENT, 0);
	const hkxVertexDescription::ElementDecl* texDecl =
		vertexDesc.getElementDecl(hkxVertexDescription::HKX_DU_TEXCOORD, 0);

	// Get the strides for each part of the vertex
	hkUint32 posStride = posDecl->m_byteStride;
	hkUint32 normStride = normDecl->m_byteStride;
	hkUint32 tanStride = 0, texStride = 0;
	if( tanDecl )
		tanDecl->m_byteStride;
	if( texStride )
		texStride = texDecl->m_byteStride;
	hkUint32 vertexStride = posStride + normStride + tanStride + texStride;

	// Now, get the base address of position, normal, tangent, and texture
	const uchar* posBase = static_cast<const uchar*>(vb->getVertexDataPtr(*posDecl));
	const uchar* normBase = static_cast<const uchar*>(vb->getVertexDataPtr(*normDecl));
	//const uchar* tanBase = static_cast<const uchar*>(vb->getVertexDataPtr(*tanDecl));
	//const uchar* texBase = static_cast<const uchar*>(vb->getVertexDataPtr(*texDecl));

	// Create the data array for the indices
	DWORD numIndices = ib->m_length;
	DxIndex* indexPtr = new DxIndex[numIndices];
	DxIndex* curIndex = indexPtr;
	int numTriangles = ib->getNumTriangles();

	// Now, go through each vertex
	for( int currentTriangle = 0; currentTriangle < numTriangles; ++currentTriangle)
	{
		// The triangle values
		hkUint32 triIndices[3];

		// Get the triangle values
		ib->getTriangleIndices(currentTriangle,	triIndices[0], triIndices[1], triIndices[2]);

		// Now, set the current vertex values
		curIndex[0].index = triIndices[0];
		curIndex[1].index = triIndices[1];
		curIndex[2].index = triIndices[2];

		// Increment the pointer
		curIndex += 3;
	}

	// Create the correct data array for the vertices
	int numVerts = vb->getNumVertices();
	DxVertex* vertexData = new DxVertex[numVerts];

	// Now, go through each vertex
	for(int currentVertex = 0; currentVertex < numVerts; ++currentVertex)
	{
		// Get the vertex data for each point
		const float* position = reinterpret_cast<const float*>(posBase + (posStride * currentVertex));
		const float* normal = reinterpret_cast<const float*>(normBase + (normStride * currentVertex));
		//const float* tangent = reinterpret_cast<const float*>(tanBase + (tanStride * currentVertex));
		//const float* texture = reinterpret_cast<const float*>(texBase + (texStride * currentVertex));

		//pos worldTransform  tempV ̵, ȸѴ.
		hkVector4 tempV;
		hkVector4 pos = hkVector4( position[0], position[1], position[2] );
		worldTransform.transformPosition(pos, tempV);

		// Set the vertex information
		//vertexData[currentVertex].position = D3DXVECTOR3(position[0], position[1], position[2]);
		vertexData[currentVertex].position = D3DXVECTOR3(tempV(0), tempV(1), tempV(2));
		vertexData[currentVertex].normal = D3DXVECTOR3(normal[0], normal[1], normal[2]);
		//vertexData[currentVertex].tangent = D3DXVECTOR3(tangent[0], tangent[1], tangent[2]);
		//vertexData[currentVertex].texcoord = D3DXVECTOR2(texture[0], texture[1]);

		// Get the min and max values for the mesh
		if(position[0] < minX)
			minX = position[0];
		else if(position[0] > maxX)
			maxX = position[0];
		if(position[1] < minY)
			minY = position[1];
		else if(position[1] > maxY)
			maxY = position[1];
		if(position[2] < minZ)
			minZ = position[2];
		else if(position[2] > maxZ)
			maxZ = position[2];
	}

	//Create DxMesh
	DxMesh* dxMesh = new DxMesh;
	dxMesh->SetIndexBuffer( numIndices, indexPtr );
	dxMesh->SetVertexBuffer( numVerts, vertexData );
	
	MeshInfo info;
	info.dxMesh = dxMesh;
	info.refRigidBody = NULL;
	m_meshArr.push_back( info );
}
//-----------------------------------------------------------------------------
void HavokDebug::Render()
{
	int nSize = (int)m_meshArr.size();
	for( int n = 0; n < nSize; ++n )
	{
		const MeshInfo& info = m_meshArr[n];
		DxMesh* pMesh = info.dxMesh;
		hkpRigidBody* pRigidBody = info.refRigidBody;

		if( pRigidBody )
		{
			const hkTransform &t = pRigidBody->getTransform();
			D3DXMATRIX transMat(
				t(0,0), t(1,0), t(2,0), 0.0f,
				t(0,1), t(1,1), t(2,1), 0.0f,
				t(0,2), t(1,2), t(2,2), 0.0f,
				t(0,3), t(1,3), t(2,3), 1.0f);
			pMesh->SetTransform( transMat );
		}

		pMesh->Render();
	}	
}
//-----------------------------------------------------------------------------